import sys
import json
from argparse import Namespace

# https://mikefarah.gitbook.io/yq/usage/convert
# yq eval -o=j mindmap.yml | python mindmap.py

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

def head(l):
    for x in l:
        return x

def merge_dicts(*ds):
    response = {}
    for d in ds:
        if isinstance(d, dict):
            response.update(d)
    return response

def merge_lists(*ls):
    response = []
    for l in ls:
        if isinstance(l, list):
            response = response + l
    return response

def recursive_print(obj, prefix="+", depth=2, style=''):
    # If its a string ...
    if isinstance(obj, str):
        print(f"{prefix*depth}{style} {obj}")
        return

    # If its a link obj
    if (
        isinstance(obj, dict) and 
        (len(obj.values()) == 1) and 
        (isinstance(head(obj.values()), str)) and 
        (head(obj.values()).startswith("http"))
    ):
        print(f"{prefix*depth}{style} [[{head(obj.values())} {head(obj.keys())}]]")
        return
    
    # If its a list ... 
    if isinstance(obj, list):
        for v in obj:
            recursive_print(v, prefix=prefix, depth=depth, style='_')
        return
        
    # Is a category or the root of a tool description
    if isinstance(obj, dict):
        # and all([isinstance(x, dict) for x in obj.values()])
        value_dict = merge_dicts(*merge_lists(*obj.values()))
        # eprint(*obj.values())
        # eprint(merge_lists(*obj.values()))
        # eprint(value_dict)
        for k, v in obj.items():
            node_class = "<<tool>>" if (("🖥" in value_dict) or ("📖" in value_dict)) else ""
            # Skip inner dicts ...
            if k in ["Uses Include", "Notes", "Questions"]:
                continue
            # if ("🖥" in value_dict) and ("📖" in value_dict):
            #     print(f"{prefix*depth}{style}: {k}\n[[{value_dict['🖥']} 🖥]]  |  [[{value_dict['📖']} 📖]] {node_class};")
            # el
            if "🖥" in value_dict:
                print(f"{prefix*depth}{style} [[{value_dict['🖥']} {k}]] {node_class}")
            else:
                print(f"{prefix*depth}{style} {k} {node_class}")
            recursive_print(v, prefix=prefix, depth=depth+1, style=style)
        return


# https://docs.python.org/3/library/json.html
# https://stackoverflow.com/questions/4984647/accessing-dict-keys-like-an-attribute
# https://docs.python.org/dev/library/types.html#types.SimpleNamespace

# JSON = json.load(sys.stdin, object_hook=lambda o: Namespace(**o))
JSON = json.load(sys.stdin)

# print(json.dumps(JSON))

print('@startmindmap')
print("""
<style>
    mindmapDiagram {
        .tool {
            BackgroundColor lightblue
        }
    }
</style>
""")
print(f'+ **{JSON["Title"]}**')
# print(json.dumps(JSON["Right"], indent=4))
recursive_print(JSON["Right"])
print("'tag::left[]")
# print(json.dumps(JSON["Left"], indent=4))
recursive_print(JSON["Left"], "-")
print("'end::left[]")
print('@endmindmap')



#!/usr/bin/env python3.7

import json
import sys

import typer

from typing import Optional, List
from pathlib import Path

from demopy.puml import synchro
from demopy.tools import clean_multiline_str
from demopy.constants import DiagramType, DEFAULTS
from demopy.puml import simple as simple_puml

app = typer.Typer()

def _generate_url(label:str, puml: str, base_url, diagram_type):
    content = synchro.compress(puml)
    return f'[{label}]({base_url}/{diagram_type}{content})'


# "https://typer.tiangolo.com/tutorial/multiple-values/multiple-options/"
@app.command()
def simple(
        chain: Optional[List[str]] = typer.Option([]), 
        resource_definitions:Optional[Path]=typer.Option(None), 
        base_url:str=typer.Option(DEFAULTS["url"]), 
        diagram_type:DiagramType=typer.Option(DEFAULTS["type"])
    ):
    """
    Simple English -> PlantUML Url.
    """
    if resource_definitions:
        with open(resource_definitions) as fh:
            custom_resources = json.load(fh)
    else:
        custom_resources = {}
    typer.echo(
        _generate_url(
            "diagram",
            simple_puml.convert_to_puml("\n".join(chain), custom_resources),
            base_url,
            diagram_type
        )
    )


@app.command()
def generate_url(path:Path, base_url:str=DEFAULTS["url"], diagram_type:DiagramType=DEFAULTS["type"]):
    """
    File w/ Puml (or stdin) -> PlantUML Url.
    """
    if path.name != "-" and not path.is_file():
        typer.echo("Path must be a file.")
        raise typer.Abort()
    if path.name == "-":
        content = sys.stdin.read().strip()
    else:
        with open(path, "r") as fh:
            content = fh.read().strip()
    typer.echo(_generate_url(path.name, content, base_url, diagram_type.value))


if __name__ == "__main__":
    app()

# http://codaset.com/repo/python-markdown/tickets/new


from typing import get_type_hints, _GenericAlias, List, Optional


def delimit_camel_case(s):
    # https://stackoverflow.com/questions/5020906/python-convert-camel-case-to-space-delimited-using-regex-and-taking-acronyms-in
    return ''.join(' ' + c if c.isupper() and s[max(0, i-1)].islower() else c for i, c in enumerate(s))

def clean_type(t):
    # https://www.delftstack.com/howto/python/python-get-class-name/
    t = str(t) if isinstance(t, _GenericAlias) else t.__name__
    return t.replace("typing.List", "").replace("__main__.", "").replace("typing.Optional", "")

def convert_class_to_puml(cls):
    name = cls.__name__
    private_fields, public_fields = convert_fields_to_puml(cls)
    private_fields, public_fields = "\n".join(private_fields), "\n".join(public_fields)
    return (f"""
    class "{delimit_camel_case(name)}" as {name} {{
        {private_fields}
        {public_fields}
    }}
    """)

def convert_fields_to_puml(cls):
    # https://docs.python.org/3/library/typing.html#introspection-helpers
    fields = get_type_hints(cls)
    private_fields = [
        f"- {clean_type(v)} {k}"
        for k, v in fields.items()
        if k.startswith("_")
    ]
    public_fields = [
        f"+ {clean_type(v)} {k}"
        for k, v in fields.items()
        if not k.startswith("_")
    ]
    return private_fields, public_fields


def base_type(cls):
    return clean_type(cls).replace("[", "").replace("]", "")

def parent_class(cls):
    return cls.__mro__[1:][0]

def convert_relationships_to_puml(cls, other_custom_resources):
    # https://docs.python.org/3/library/typing.html#introspection-helpers
    relationships = []
    fields = get_type_hints(cls)
    for k, v in fields.items():
        # print(k, v, clean_type(cls), clean_type(v), base_type(v))
        if base_type(v) in other_custom_resources:
            if clean_type(v) == f'[{base_type(v)}]':
                rel = f"{base_type(cls)} --{{ {base_type(v)}"
            else:
                rel = f"{base_type(cls)} -- {base_type(v)}"
        else:
            continue 
        relationships = relationships + [rel]
    if base_type(parent_class(cls)) in other_custom_resources:
        relationships = relationships + [
            f'{base_type(cls)} --o {base_type(parent_class(cls))}'
        ]
    return "\n".join(relationships)


def header():
    return """
@startuml
' skinparam linetype ortho
  skinparam packageStyle rectangle
  skinparam shadowing false
  skinparam class {
    BackgroundColor White
    BorderColor Black
    ArrowColor Black
  }
  hide circle
"""


def footer():
    return "@enduml"


#   class "User" as User1 {
#     - int id
#     - int age
#     - string name
#     + ([role]) roles()
#     + ([user_role]) user_roles()
#   }
#   class Role {
#     - int id
#     - string name
#   }
#   class "UserRole" as User1Role {
#     - int id
#     - int user_id
#     - int role_id
#   }



# Python implementation of: 
#   https://plantuml.com/synchro2.min.js
# See also ... 
#   * encoded-entry.html
#   * pako.js

# https://careerkarma.com/blog/python-indexerror-list-assignment-index-out-of-range/
# https://stackoverflow.com/questions/5653533/why-does-this-iterative-list-growing-code-give-indexerror-list-assignment-index


import zlib


def generate6BitEncodingMap():
    r = {}
    for i in range(0, 64):
        if (i < 10):
            # digits
            r[i] = chr(0x30 + i); 
        elif (i < 36):
            # uppercase letters
            r[i] = chr(0x41 + (i - 10))
        elif (i < 62):
            # lowercase letters
            r[i] = chr(0x61 + (i - 36))
        elif (i == 62):
            r[i] = '-'
        else: 
            r[i] = '_'
    return r


def generate6BitDecodingMap(encodingMap):
    return {
        e: i
        for i, e in encodingMap.items()
    }


SIXBIT_ENCODING_MAP = generate6BitEncodingMap()
SIXBIT_DECODING_MAP = generate6BitDecodingMap(SIXBIT_ENCODING_MAP)


def append3bytes(b1, b2, b3):
    c1 = b1 >> 2
    c2 = ((b1 & 0x3) << 4) | (b2 >> 4)
    c3 = ((b2 & 0xF) << 2) | (b3 >> 6)
    c4 = b3 & 0x3F
    r = ""
    r += SIXBIT_ENCODING_MAP[c1 & 0x3F]
    r += SIXBIT_ENCODING_MAP[c2 & 0x3F]
    r += SIXBIT_ENCODING_MAP[c3 & 0x3F]
    r += SIXBIT_ENCODING_MAP[c4 & 0x3F]
    return r


def decode3bytes(r, cc1, cc2, cc3, cc4):
    c1 = SIXBIT_DECODING_MAP[cc1]
    c2 = SIXBIT_DECODING_MAP[cc2]
    c3 = SIXBIT_DECODING_MAP[cc3]
    c4 = SIXBIT_DECODING_MAP[cc4]
    yield ((c1 << 2) | (c2 >> 4))
    yield (((c2 & 0x0F) << 4) | (c3 >> 2))
    yield (((c3 & 0x3) << 6) | c4)


def encode(data):
    # print(type(data), data, data[2])
    r = ""
    for i in range(0, len(data), 3):
        if (i + 2 == len(data)):
            r += append3bytes(data[i], data[i + 1], 0)
        elif (i + 1 == len(data)):
            r += append3bytes(data[i], 0, 0)
        else:
            r += append3bytes(data[i], data[i + 1], data[i + 2])
    return r


def decode(s):
    if not ( (len(s) % 4) == 0 ):
        raise Exception("Failed decoding: [" + s + "]")
    r = []
    for i in range(0, len(s), 4):
        for b in decode3bytes(
            r,
            s[i],
            s[i + 1],
            s[i + 2],
            s[i + 3]
        ):
            r.append(b)
    # print(r)
    # return "".join([str(x) for x in r])
    return bytearray(r)


def compress(raw_str):
    # print("input::raw_str", raw_str)
    compressed_bytesarry = zlib.compress(raw_str.encode('utf-8'), 9)
    # print("\tcompressed_bytesarry", compressed_bytesarry)
    encoded_bytearray = encode(compressed_bytesarry)
    # print("output::encoded_bytearray", encoded_bytearray)
    return encoded_bytearray


def decompress(encoded_bytearray):
    # print("input::encoded_bytearray", encoded_bytearray)
    decoded_bytesarray = decode(encoded_bytearray)
    # print("\tdecoded_bytesarray", decoded_bytesarray)
    decompressed_string = zlib.decompress(decoded_bytesarray).decode('utf-8')
    # print("output::decompressed_string", decompressed_string)
    return decompressed_string


# print(SIXBIT_ENCODING_MAP)
# print(SIXBIT_DECODING_MAP)
# print("compressed: ", c)
# print("decompressed: ", decompress(c))



from collections import namedtuple, defaultdict
from itertools import combinations
import re
from typing import List
from jinja2 import Template
# https://realpython.com/primer-on-jinja-templating/
from demopy.tools import clean_multiline_str


Relationship = namedtuple('Relationship', ['src', 'dst', 'rel', 'num'])


def parse_rels_from_line(simple_statement:str, line_number:int=1):
    lines = re.sub(r"\]->([^-]+)-\[", r"]->\1\n\1-[", simple_statement).split("\n")
    return [
        eval(re.sub(
            r"(.*)-\[(.*)\]->(.*)", 
            f'Relationship("\\1".strip(), "\\3".strip(), "\\2".strip(), {line_number})', 
            line
        ))
        for line in lines
    ]


def convert_to_puml(simple_statement:str, resources:dict={}) -> str:
    simple_statement = simple_statement.replace("\\n", "\n")
    rels = [
        rel
        for i, l in enumerate(simple_statement.split("\n"))
        for rel in parse_rels_from_line(l, i+1)
    ]
    # print(rels)
    return clean_multiline_str(
        "\n".join([
            header(),
            define_resources_involved_in_rels(rels, resources),
            define_rels(rels),
            define_same_links(rels),
            footer()
        ])
    )


def header() -> str:
    return """
    @startuml
    !theme cerulean
    left to right direction
    
    ' https://plantuml.com/stdlib
    ' https://github.com/awslabs/aws-icons-for-plantuml#basic-usage
    
    !define AWSPuml https://raw.githubusercontent.com/awslabs/aws-icons-for-plantuml/v11.1/dist
    !includeurl AWSPuml/AWSCommon.puml
    !includeurl AWSPuml/AWSSimplified.puml
    !includeurl AWSPuml/ApplicationIntegration/APIGateway.puml
    !includeurl AWSPuml/Compute/Lambda.puml
    !includeurl AWSPuml/Database/DynamoDB.puml
    !includeurl AWSPuml/General/Users.puml
    !includeurl AWSPuml/SecurityIdentityCompliance/Cognito.puml
    !includeurl AWSPuml/Storage/SimpleStorageService.puml

    ' https://plantuml.com/style-evolution
    'skinparam linetype polyline
    'skinparam linetype ortho
    skinparam package<<Layout>> {
        BorderColor Transparent
        BackgroundColor Transparent
        FontColor Transparent
        TitleFontColor Transparent
        StereotypeFontColor Transparent
    }
    """
# TODO ... support bidirectional arrows ...

def footer()->str:
    return "@enduml"

    
def determine_node_repitition(rels:List[Relationship]):
    lines_per_node = defaultdict(set)
    for rel in rels:
        lines_per_node[rel.src].add(rel.num)
        lines_per_node[rel.dst].add(rel.num)
    return lines_per_node


def define_rels(rels:List[Relationship]) -> str:
    t = Template(
    """
    {% for rel in rels %}
    {{rel.src}}{{rel.num}} --> {{rel.dst}}{{rel.num}} : {{rel.rel}}
    {% endfor %}
    """
    )
    return t.render(rels=rels)


def define_same_links(rels:List[Relationship]) -> str:
    lines_per_node = determine_node_repitition(rels)
    combinations_per_node = {k: list(combinations(v, 2)) for k, v in lines_per_node.items()}
    t = Template(
    """
    {% for (node, combinations) in combinations_per_node.items() %}
    {% for (left, right) in combinations %}
    {{node}}{{left}} ..right.. {{node}}{{right}}: Same
    {% endfor %}
    {% endfor %}
    """
    )
    return t.render(combinations_per_node=combinations_per_node)


def define_resource(id, line, resources):
    awsdef = {x["id"]: x for x in resources.get("aws", [])}.get(id)
    if awsdef:
        awsdef["tooltip"] = awsdef["title"] if not awsdef.get("tooltip") else awsdef["tooltip"]
        awsdef["title"] = (
            awsdef["title"]
            # Use title if link not defined or already expanded into link ...
            if (not awsdef.get("link") or awsdef.get("title").startswith("[[")) 
            else "[[{link}{{{tooltip}}} {title}]]".format(**awsdef)
        )
        #  { "id": "sources", "type": "Users", "title": "Events", "note": "millions of users" },
        return '{type}({id}{line}, "{title}", "{note}")'.format(**awsdef, line=line)
    return f"rectangle {id}{line}"


def define_resources_involved_in_rels(rels:List[Relationship], resources:dict):
    lines_per_node = determine_node_repitition(rels)
    t = Template(
    """
    {% for node, lines in lines_per_node.items() %}
    package " " as {{node}} <<Layout>> {
        {% for line in lines %}
         {{ define_resource(node, line, resources) }} 
        {% endfor %}
    }
    {% endfor %}
    """
    )
    # https://stackoverflow.com/questions/6036082/call-a-python-function-from-jinja2
    t.globals["define_resource"] = define_resource
    t.globals["resources"] = resources
    return t.render(lines_per_node=lines_per_node)


if __name__ == "__main__":
    print(convert_to_puml("a-[knows]->b-[knows]->c"))
    print(convert_to_puml("a -[knows]->b-[knows]->c\n   a     -[owes]->c"))
    print(convert_to_puml(
        """
        """, 
    ))


# -*- coding: utf8 -*-

from setuptools import setup, find_packages
from setuptools.command.install import install
from setuptools.command.develop import develop
from subprocess import check_call

# https: // stackoverflow.com/questions/20288711/post-install-script-with-python-setuptools

# Todo ... figure out how to only do these ... if the notebooks extras is installed ...

# class CustomDevelopCommand(develop):
#     """Custom Develop Command"""

#     def run(self):
#         develop.run(self)
#         print("Installing jupyter extension")
#         check_call("jupyter nbextension enable --py widgetsnbextension".split())


# class CustomInstallCommand(install):
#     """Custom Installation Command"""

#     def run(self):
#         install.run(self)
#         check_call("jupyter nbextension enable --py widgetsnbextension".split())


with open("README.md") as fh:
    long_description = fh.read()

setup(
    # Basic info
    name="demo-py",
    version="0.0.1",
    author='Omar Eid',
    author_email='contact.omar.eid@gmail.com',
    url='https://github.com/omars-lab/demo-py',
    description='My Personal Demo Toolbox.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries',
    ],
    # cmdclass={
    #     "develop": CustomDevelopCommand,
    #     'install': CustomInstallCommand,
    # },
    # Packages and depencies
    packages=find_packages(include=["demopy", "demopy.*"]),
    install_requires=[
        "typer",        
        "pandas",
    ],
    extras_require={
        "notebooks": [
            "matplotlib",
            "ipython",
            "ipywidgets",
            "jupyter_contrib_nbextensions",
            "spacy >= 2.3.2", # eventually move this over ... too parsing extras
        ],
        'parsing': [
            'lark-parser'
            "spacy >= 2.3.2",
        ],
        'dev': [
            "mypy",
            "pytest",
            "pytest-pep8",
            "pytest-profiling",  # https://pypi.org/project/pytest-profiling/
            "pstats-view",  # https://github.com/ssanderson/pstats-view
            "flake8",
            # 'manuel',
            # 'pytest-cov',
            # 'coverage',
            # 'mock',
            # "nbconvert==5.5.0",
        ],
        "graph": [
            "py2cytoscape",
            "pycandela",
            "pygraphviz",
        ],
        "ext": [
            "mpldatacursor",  #(https://github.com/joferkington/mpldatacursor)
            "calmap",  #(https://github.com/martijnvermaat/calmap/blob/master/calmap/__init__.py),
            "ipython[notebook]",
        ],
        "toolz": [
            "attrs",
            "pydantic",
            "toolz",  # https://toolz.readthedocs.io/en/latest/api.html
            "numpy",
            # 'Jinja2', # https://jinja.palletsprojects.com/en/3.0.x/intro/
        ]
    },
    # Data files
    package_data={},
    # Scripts
    entry_points={
        'console_scripts': [
            'my-puml = demopy.cli.puml:app'],
    },
    # Other configurations
    zip_safe=False,
    platforms='any'
)



from enum import Enum


class DiagramType(str, Enum):
    huffman_svg = "svg/~1"
    huffman_editor = "uml/~1"
    huffman_uml = "uml/~1"
    huffman_png = "png/~1"


DEFAULTS = {
    "url": "http://www.plantuml.com/plantuml",
    "type": DiagramType.huffman_uml
}